/**
 * gsvnc_sps2.c by Sauce (Steven Osman) & Sparky (Morten Mikkelsen)
 *
 * This file has the implementation for the sps2 portion of the gsvnc package.
 * The functions in this file and its gsvnc dependencies (currently only
 * gsvnc.h and gsvnc_sps2.h) may be used freely so long as the sps2 license is
 * adhered to.
 *
 * Please refer to gsvnc_sps2.h for usage information
 */

#include "gsvnc.h"
#include "sps2registers.h"
#include "sps2types.h"
#include "sps2lib.h"
#include <sys/ipc.h>
#include <sys/shm.h>
#include <string.h>
#include <linux/ps2/dev.h>
#include <linux/ps2/gs.h>

void _gsvncDoUpdate(int iWidth, int iHeight,
		    unsigned int uiFirstBuffer, unsigned int uiSecondBuffer,
		    unsigned int uiZBuffer,
		    unsigned int uiRenderBuffer,
		    unsigned int uiDisplayBuffer);
void _gsvncStartUpXFer(void *virtPage, int physPage, int addr, int bufw, 
		       int x, int y, int w, int h, int iSourcePixelFormat);
void _gsvncStartCopyScanLine(unsigned int physPage, int w, int h);
void _gsvncEndCopyScanLine();
void _gsvncShutDownXFer();
void _gsvncBackupSpr();
void _gsvncRestoreSpr();
void _gsvncConvRGBA_1(void * pColorsOut, const void * pColorsIn, const sps2uint32 iNumCacheLinesIn);
void _gsvncConvRGBA_2(void * pColorsOut, const void* pColorsIn, const sps2uint32 iNumCacheLinesIn);

int _iGSVNCSPS2Handle=-1;
sps2Memory_t *_pGSVNCSPS2CachedDescriptor=0;
sps2Memory_t *_pGSVNCSPS2UncachedDescriptor=0;
frameBufferData *_pGSVNCFrameBuffer=0;
void *_pvGSVNCCachedBuffer[2];
void *_pvGSVNCUncachedBuffer[2];
sps2uint32 _uiGSVNCBufferPhysAddress[2];
int _iGSVNCLastFrameSent=0;
gsvncBufferTypes_t _iGSVNCLastBufferSent=(gsvncBufferTypes_t) -1;
int _iGSVNCLastPageSent=-1;
int _iGSVNCLastFrameClamped=0;
int _iGSVNCLastFrameClampedBits=0;
int _iGSVNCUseScratchPad=1;
void *_pvGSVNCEEBase;
void *_pvGSVNCGSBase;
void *_pvGSVNCScratchPad;
void *_pvGSVNCScratchPadBuffer[2];
sps2uint32 _uiGSVNCScratchPadPhysAddress[]={1 << 31, (1 << 31) | 4096};

int gsvncSPS2Init(int iUseScratchPad) {
  if ((_pGSVNCFrameBuffer=_gsvncGetSharedFrameBuffer())==0) {
    return -1;
  }

  _iGSVNCSPS2Handle=_sps2Open();

  if (_iGSVNCSPS2Handle<0) {
    return -2;
  }

  _pGSVNCSPS2UncachedDescriptor=sps2Allocate(8192,
				 SPS2_MAP_BLOCK_4K | SPS2_MAP_UNCACHED,
				 _iGSVNCSPS2Handle);

  if (_pGSVNCSPS2UncachedDescriptor==0) {
    return -2;
  }

  _pvGSVNCEEBase=_sps2MapEERegisters(0, _iGSVNCSPS2Handle);
  _pvGSVNCGSBase=_sps2MapGSRegisters(0, _iGSVNCSPS2Handle);
  _sps2SetEIDIEnabled(1, _iGSVNCSPS2Handle);

  _pGSVNCSPS2CachedDescriptor=sps2Remap(_pGSVNCSPS2UncachedDescriptor,
					SPS2_MAP_BLOCK_4K | SPS2_MAP_CACHED,
					_iGSVNCSPS2Handle);

  _pvGSVNCUncachedBuffer[0]=_pGSVNCSPS2UncachedDescriptor->pvStart;
  _pvGSVNCUncachedBuffer[1]=(void *) 
    (((char *)_pGSVNCSPS2UncachedDescriptor->pvStart)+4096);

  _pvGSVNCCachedBuffer[0]=_pGSVNCSPS2CachedDescriptor->pvStart;
  _pvGSVNCCachedBuffer[1]=(void *)
    (((char *) _pGSVNCSPS2CachedDescriptor->pvStart)+4096);

  _uiGSVNCBufferPhysAddress[0]=
    sps2GetPhysicalAddress(_pvGSVNCUncachedBuffer[0],
			   _pGSVNCSPS2UncachedDescriptor);
  _uiGSVNCBufferPhysAddress[1]=
    sps2GetPhysicalAddress(_pvGSVNCUncachedBuffer[1],
			   _pGSVNCSPS2UncachedDescriptor);

  _iGSVNCUseScratchPad=iUseScratchPad;

  if (_iGSVNCUseScratchPad) {
    _pvGSVNCScratchPad=_sps2MapScratchPad(0, _iGSVNCSPS2Handle);
    _pvGSVNCScratchPadBuffer[0]=_pvGSVNCScratchPad;
    _pvGSVNCScratchPadBuffer[1]=(void *) (((char *) _pvGSVNCScratchPad)+4096);
  }


  return 0;
}

int gsvncSPS2Update(int iWidth, int iHeight, int iUpdateRate, 
		    unsigned int uiFirstBuffer, unsigned int uiSecondBuffer,
		    unsigned int uiZBuffer,
		    unsigned int uiRenderBuffer,
		    unsigned int uiDisplayBuffer) {
  int iFrameDirty, iFirstPass=0;

  if (!_pvGSVNCUncachedBuffer) {
    return -1;
  }

  if (!_pGSVNCFrameBuffer->iClientReferenceCounter) {
    return 1;
  }

  _iGSVNCLastFrameSent++;

  do {
    iFrameDirty=0;

    if (iFirstPass) {
      iFirstPass=0;

      // Make sure that if the system is paused, we update the buffer to be
      // in sync with what is on the screen.  Do this only the first time we
      // enter this loop.
      if (_pGSVNCFrameBuffer->iSystemPaused) {
        iFrameDirty=1;
      }
    }

    // If capture is not paused, see if enough frames have passed that we
    // need to update
    if (_iGSVNCLastFrameSent>=iUpdateRate &&
	!_pGSVNCFrameBuffer->iCapturePaused) {
      iFrameDirty=1;
    } else if (_iGSVNCLastBufferSent!=_pGSVNCFrameBuffer->iRequestedBuffer) {
      // Also, see if the last buffer we sent is the most recent one requested
      iFrameDirty=1;
    } else if (_iGSVNCLastBufferSent==GSVNC_BUFFER_ANY_PAGE) {
      if (_iGSVNCLastPageSent!=_pGSVNCFrameBuffer->iRequestedPage) {
	iFrameDirty=1;
      }
    } else if (_iGSVNCLastBufferSent==GSVNC_BUFFER_Z_BUFFER) {
      // If we're doing Z buffer, also see if the clamped format was consistent
      // with the last clamped format we sent, both based on whether it is on
      // or off and whether we clamped the same range of bits.
      if (_iGSVNCLastFrameClamped!=_pGSVNCFrameBuffer->iZBufferClamping) {
	iFrameDirty=1;
      } else if (_iGSVNCLastFrameClamped && 
		 _iGSVNCLastFrameClampedBits!=_pGSVNCFrameBuffer->iClampBits) {
	iFrameDirty=1;
      }
    }

    if (iFrameDirty) {
      _iGSVNCLastFrameSent=0;

      _gsvncDoUpdate(iWidth, iHeight, uiFirstBuffer, uiSecondBuffer,
		     uiZBuffer, uiRenderBuffer, uiDisplayBuffer);
    }
  } while (_pGSVNCFrameBuffer->iSystemPaused &&
	   _pGSVNCFrameBuffer->iClientReferenceCounter);

  return 0;
}

void _gsvncDoUpdate(int iWidth, int iHeight,
		    unsigned int uiFirstBuffer, unsigned int uiSecondBuffer,
		    unsigned int uiZBuffer,
		    unsigned int uiRenderBuffer,
		    unsigned int uiDisplayBuffer) {
  int iYLooper, iXLooper;
  unsigned int uiSourceBuffer=uiFirstBuffer;
  int iSourcePixelFormat=0;
  unsigned char *pucSource, *pucDestination;
  int iLoadBuffer;
  int iFirstClampBit=_pGSVNCFrameBuffer->iClampBits;
  sps2uint32 uiBackupMADR, uiBackupQWC, uiBackupCHCR;

  // vif1 must be available
  if ((*EE_D1_CHCR_OFF(_pvGSVNCEEBase) & 0x0100) != 0) {
    return;
  }	

  _iGSVNCLastBufferSent=_pGSVNCFrameBuffer->iRequestedBuffer;
  _iGSVNCLastFrameClamped=_pGSVNCFrameBuffer->iZBufferClamping;
  _iGSVNCLastFrameClampedBits=_pGSVNCFrameBuffer->iClampBits;
  _iGSVNCLastPageSent=_pGSVNCFrameBuffer->iRequestedPage;

  switch (_pGSVNCFrameBuffer->iRequestedBuffer) {
  case GSVNC_BUFFER_FIRST:
    uiSourceBuffer=uiFirstBuffer;
    iSourcePixelFormat=0;
    break;
  case GSVNC_BUFFER_SECOND:
    uiSourceBuffer=uiSecondBuffer;
    iSourcePixelFormat=0;
    break;
  case GSVNC_BUFFER_Z_BUFFER:
    uiSourceBuffer=uiZBuffer;
    iSourcePixelFormat=0x30;
    break;
  case GSVNC_BUFFER_RENDER:
    uiSourceBuffer=uiRenderBuffer;
    iSourcePixelFormat=0;
    break;
  case GSVNC_BUFFER_DISPLAY:
    uiSourceBuffer=uiDisplayBuffer;
    iSourcePixelFormat=0;
    break;
  case GSVNC_BUFFER_ANY_PAGE:
    uiSourceBuffer=_pGSVNCFrameBuffer->iRequestedPage;
    iSourcePixelFormat=0;
    break;    
  default:
    printf("Unexpected buffer specified %i!!!\n", 
	   _pGSVNCFrameBuffer->iRequestedBuffer);
  }

  pucDestination=_pGSVNCFrameBuffer->aucFrameBufferDataArea;
  GSVNC_ALIGN_POINTER(pucDestination,16);

  uiBackupMADR=*EE_D1_MADR_OFF(_pvGSVNCEEBase);
  uiBackupQWC=*EE_D1_QWC_OFF(_pvGSVNCEEBase);
  uiBackupCHCR=*EE_D1_CHCR_OFF(_pvGSVNCEEBase);

  _gsvncStartUpXFer(_pvGSVNCUncachedBuffer[0], _uiGSVNCBufferPhysAddress[0],
		    32 * uiSourceBuffer, (iWidth + 63)/64, 0,
		    0, iWidth, iHeight, iSourcePixelFormat);

  if (_iGSVNCUseScratchPad) {
    // Back up spr _after_ starting transfer
    // This is because we share the same memory for spr backup as we do for
    // the transfer.
    _gsvncBackupSpr();
  }

  iLoadBuffer=0;

  // Start the first transfer
  if (_iGSVNCUseScratchPad) {
    _gsvncStartCopyScanLine(_uiGSVNCScratchPadPhysAddress[iLoadBuffer], iWidth,
			    1);
  } else {
    _gsvncStartCopyScanLine(_uiGSVNCBufferPhysAddress[iLoadBuffer], iWidth, 1);
  }

  for(iYLooper = 0; iYLooper < iHeight; iYLooper++) {
    iLoadBuffer=1-iLoadBuffer;

    // End the current transfer
    _gsvncEndCopyScanLine();

    if (!_iGSVNCUseScratchPad) {
      sps2FlushCache(_iGSVNCSPS2Handle);
    }

    // Start the next transfer if we're not at the last row
    if (iYLooper+1<iHeight) {
      if (_iGSVNCUseScratchPad) {
	_gsvncStartCopyScanLine(_uiGSVNCScratchPadPhysAddress[iLoadBuffer],
				iWidth, 1);
      } else {
	_gsvncStartCopyScanLine(_uiGSVNCBufferPhysAddress[iLoadBuffer],
				iWidth, 1);
      }
    }

    // Process the current transfer
    if (_iGSVNCUseScratchPad) {
      pucSource=(unsigned char *) _pvGSVNCScratchPadBuffer[1-iLoadBuffer];
    } else {
      pucSource=(unsigned char *) _pvGSVNCCachedBuffer[1-iLoadBuffer];
    }

    if (_pGSVNCFrameBuffer->iRequestedBuffer!=GSVNC_BUFFER_Z_BUFFER ||
	_pGSVNCFrameBuffer->iZBufferClamping==0) {

      // If we have a number of colors that is divisible by 16, we can
      // use the super-fast copying code
      if (iWidth % 16==0) {
	_gsvncConvRGBA_2(pucDestination, pucSource, iWidth/16);
	pucDestination+=iWidth*3;
      } else {
	// Copy only R, G and B values using slow version of the code
	for (iXLooper=0; iXLooper<iWidth; iXLooper++) {
	  asm __volatile__ ("
             pref 0, 16(%0)
          " : : "r" (pucSource));
	  
	  *(pucDestination++)=*pucSource++;
	  *(pucDestination++)=*pucSource++;
	  *(pucDestination++)=*pucSource++;
	  pucSource++;
	}
      }
    } else {
      // Copy only R, G and B values
      for (iXLooper=0; iXLooper<iWidth; iXLooper++) {
	sps2uint32 uiDepth=(*(sps2uint32 *) pucSource) & 0xffffff;
	unsigned char ucClamp;

	asm __volatile__ ("
         pref 0, 16(%0)
	 " : : "r" (pucSource));

	uiDepth>>=iFirstClampBit;

	if (uiDepth >= 0xff) {
	  ucClamp=0xff;
	} else {
	  ucClamp=(unsigned char) uiDepth;
	}

	*(pucDestination++)=ucClamp;
	*(pucDestination++)=ucClamp;
	*(pucDestination++)=ucClamp;

	pucSource+=4;
      }
    }

    pucDestination+=3*(640-iWidth);
  }

  _gsvncShutDownXFer();

  if (_iGSVNCUseScratchPad) {
    _gsvncRestoreSpr();
  }

  *EE_D1_MADR_OFF(_pvGSVNCEEBase)=uiBackupMADR;
  *EE_D1_QWC_OFF(_pvGSVNCEEBase)=uiBackupQWC;
  *EE_D1_CHCR_OFF(_pvGSVNCEEBase)=uiBackupCHCR;

  _pGSVNCFrameBuffer->iWidth=iWidth;
  _pGSVNCFrameBuffer->iHeight=iHeight;
  _pGSVNCFrameBuffer->iUpdated=1;
}

void _gsvncStartCopyScanLine(unsigned int physPage, int w, int h) {
  sps2uint32 uQSize = (w*h*4)/16; // assuming 24/32bit framebuffer  

  *EE_D1_QWC_OFF(_pvGSVNCEEBase) = uQSize;
  *EE_D1_MADR_OFF(_pvGSVNCEEBase) = ( sps2uint32 ) physPage & 0x8FFFFFFF ;
  *EE_D1_CHCR_OFF(_pvGSVNCEEBase) =
    0 | (0<<2) | (0<<4) | (0<<6) | (0<<7) | (1<<8);

  asm __volatile__( " sync.l " );
}

void _gsvncEndCopyScanLine() {
  // check if DMA is complete (STR=0)
  while ( *EE_D1_CHCR_OFF(_pvGSVNCEEBase) & 0x0100 );

  asm __volatile__( " sync.l " );
}

void _gsvncStartUpXFer(void *virtPage, int physPage, int addr, int bufw, 
		       int x, int y, int w, int h, int iSourcePixelFormat) {
  sps2uint32 *pDmaMem = (sps2uint32*)virtPage;
  sps2uint64 *pDmaMem64 = (sps2uint64 *) (pDmaMem + 4);
  
  void * pPhysDmaMem = (void*)physPage;
  
  sps2uint32 uSrcPtr = addr;

  // Setup transfer texture back to memory
  pDmaMem[0] = 0; //NOP;
  pDmaMem[1] = 0x06008000; //MSKPATH3(0x8000);
  pDmaMem[2] = 0x13000000; //FLUSHA;
  pDmaMem[3] = 0x50000006; //DIRECT(6);
  
  
  // Add GifTag Data
  // GIFTAG(NLOOP, EOP, PRE, PRIM, FLG, NREG)
  pDmaMem64[0] = (PS2_GIFTAG_SET_TOPHALF(5, 1, 0, 0, 0, 1)); 
  pDmaMem64[1] = (0xe);
  
  pDmaMem64[2] = (PS2_GS_SETREG_BITBLTBUF( uSrcPtr, bufw, iSourcePixelFormat,
					   0, 0, 0 ));
  pDmaMem64[3] = (PS2_GS_BITBLTBUF);
    
  // TRXPOS(SSAX, SSAY, DSAX, DSAY, DIR)
  pDmaMem64[4] = (PS2_GS_SETREG_TRXPOS(x, y, 0, 0, 0)); 
  
  pDmaMem64[5] = (PS2_GS_TRXPOS);
  
  pDmaMem64[6] = (PS2_GS_SETREG_TRXREG(w, h)); // RRW, RRh
  pDmaMem64[7] = (PS2_GS_TRXREG);
  
  pDmaMem64[8] = (0);
  pDmaMem64[9] = (PS2_GS_FINISH);
  
  pDmaMem64[10] = (PS2_GS_SETREG_TRXDIR(1)); // XDIR
  pDmaMem64[11] = (PS2_GS_TRXDIR);
  
  *EE_D1_QWC_OFF(_pvGSVNCEEBase) = 0x7;
  *EE_D1_MADR_OFF(_pvGSVNCEEBase) = ( unsigned int ) pPhysDmaMem & 0x0FFFFFFF ;

  DPUT_GS_CSR_OFF(_pvGSVNCGSBase, 2); // Enable the FINISH event

  // Disable interrupts so we don't lose our finish event
  asm __volatile__ ("di");

  *EE_D1_CHCR_OFF(_pvGSVNCEEBase) =
    (1<<0) | (0<<2) | (0<<4) | (0<<6) | (0<<7) | (1<<8);
  
  asm __volatile__( " sync.l " );
  // check if DMA is complete (STR=0)
  while ( *EE_D1_CHCR_OFF(_pvGSVNCEEBase) & 0x0100 );

  // Wait for finish event
  while ((*GS_CSR_OFF(_pvGSVNCGSBase) & 2) == 0);

  // Re-enable interrupts
  asm __volatile__ ("ei");

  // wait for VIF1-FIFO to empty and become idle
  while ( *EE_VIF1_STAT_OFF(_pvGSVNCEEBase) & 0x1f000003 );
  *((volatile sps2uint32 *) EE_VIF1_STAT_OFF(_pvGSVNCEEBase)) = (1<<23);
  DPUT_GS_BUSDIR_OFF(_pvGSVNCGSBase, ( sps2uint64 ) 0x00000001 );
}

void _gsvncShutDownXFer() {
  static sps2uint32 enable_path3[4] __attribute__ ((aligned(16))) = {
    0x06000000, //MSKPATH3(0),
    0,
    0,
    0};

  // wait for VIF1-FIFO to empty and become idle
  while ( *EE_VIF1_STAT_OFF(_pvGSVNCEEBase) & 0x1f000003 );
  *((volatile sps2uint32 *) EE_VIF1_STAT_OFF(_pvGSVNCEEBase)) = 0;
  DPUT_GS_BUSDIR_OFF(_pvGSVNCGSBase, ( sps2uint64 ) 0 );
  
  DPUT_EE_VIF1_FIFO_OFF(_pvGSVNCEEBase, *( sps2uint128 * ) enable_path3 );
}

void _gsvncBackupSpr() {
  int iLooper;
  unsigned int uiSADR, uiMADR, uiQWC, uiCHCR;
  uiSADR=*EE_D8_SADR_OFF(_pvGSVNCEEBase);
  uiMADR=*EE_D8_MADR_OFF(_pvGSVNCEEBase);
  uiQWC=*EE_D8_QWC_OFF(_pvGSVNCEEBase);
  uiCHCR=*EE_D8_CHCR_OFF(_pvGSVNCEEBase);

  for (iLooper=0;iLooper<2;iLooper++) {
    *EE_D8_SADR_OFF(_pvGSVNCEEBase)=4096 * iLooper;
    *EE_D8_MADR_OFF(_pvGSVNCEEBase)=_uiGSVNCBufferPhysAddress[iLooper];
    *EE_D8_QWC_OFF(_pvGSVNCEEBase)=4096 / 16;
    *EE_D8_CHCR_OFF(_pvGSVNCEEBase)=0x100;

    while (*EE_D8_CHCR_OFF(_pvGSVNCEEBase) & 0x100);
  }

  *EE_D8_SADR_OFF(_pvGSVNCEEBase)=uiSADR;
  *EE_D8_MADR_OFF(_pvGSVNCEEBase)=uiMADR;
  *EE_D8_QWC_OFF(_pvGSVNCEEBase)=uiQWC;
  *EE_D8_CHCR_OFF(_pvGSVNCEEBase)=uiCHCR;
}

void _gsvncRestoreSpr() {
  int iLooper;
  unsigned int uiSADR, uiMADR, uiQWC, uiCHCR;
  uiSADR=*EE_D9_SADR_OFF(_pvGSVNCEEBase);
  uiMADR=*EE_D9_MADR_OFF(_pvGSVNCEEBase);
  uiQWC=*EE_D9_QWC_OFF(_pvGSVNCEEBase);
  uiCHCR=*EE_D9_CHCR_OFF(_pvGSVNCEEBase);

  for (iLooper=0;iLooper<2;iLooper++) {
    *EE_D9_SADR_OFF(_pvGSVNCEEBase)=4096 * iLooper;
    *EE_D9_MADR_OFF(_pvGSVNCEEBase)=_uiGSVNCBufferPhysAddress[iLooper];
    *EE_D9_QWC_OFF(_pvGSVNCEEBase)=4096 / 16;
    *EE_D9_CHCR_OFF(_pvGSVNCEEBase)=0x100;

    while (*EE_D9_CHCR_OFF(_pvGSVNCEEBase) & 0x100);
  }

  *EE_D9_SADR_OFF(_pvGSVNCEEBase)=uiSADR;
  *EE_D9_MADR_OFF(_pvGSVNCEEBase)=uiMADR;
  *EE_D9_QWC_OFF(_pvGSVNCEEBase)=uiQWC;
  *EE_D9_CHCR_OFF(_pvGSVNCEEBase)=uiCHCR;
}

// cacheline is 64bytes (16 colors in rgba)
// suggestion, enable ucab for the output pointer!
void _gsvncConvRGBA_1(void * pColorsOut, const void * pColorsIn, const sps2uint32 iNumCacheLinesIn)
{
	// no pref+no ucab:			   109 cycles per itteration (output: 16 x 24bit colors)
	// pref+no ucab:				83 cycles per itteration (output: 16 x 24bit colors)
	// pref+ucab_at_destination:	57 cycles per itteration (output: 16 x 24bit colors)
	// scratchpad:					39 cycles per itteration (output: 16 x 24bit colors)
	asm __volatile__ ( "
		.set	push
		.set	mips3
		
		sll		$3, %2, 6		# mul by 64
		move	$24, %1			# src colors
		add		$3, %1, $3		# stop_pointer (offset by source pointer)
		move	$25, %0			# dst colors
		li		$8, 0x00ffffff
		pcpyld	$14, $8, $8
		dsll	$8, $8, 24
		pcpyld	$15, $8, $8
		mtsab	$0, 1

		.set	noreorder
		.align	4

	0:
		lq		$8, 0($24)
		lq		$9, 16($24)
		lq		$10, 32($24)
		lq		$11, 48($24)
		pref	0, 64*1($24)
		
		
		qfsrv	$2, $0, $8
		pand	$8, $14, $8
		pand	$2, $15, $2
		por		$8, $2, $8

		qfsrv	$2, $0, $9
		pand	$9, $14, $9
		pand	$2, $15, $2
		por		$9, $2, $9

		qfsrv	$2, $0, $10
		pand	$10, $14, $10
		pand	$2, $15, $2
		por		$10, $2, $10

		qfsrv	$2, $0, $11
		pand	$11, $14, $11
		pand	$2, $15, $2
		por		$11, $2, $11
		
		# write first quad
		sd		$8, 0 + 0*16($25)
		pcpyud	$2, $8, $0
		sh		$2, 6 + 0*16($25)
		dsrl	$2, $2, 16
		sw		$2, 8 + 0*16($25)
		sw		$9, 12 + 0*16($25)

		# write second quad
		pcpyud	$2, $9, $0
		dsrl	$9, $9, 32
		sh		$9, 0 + 1*16($25)
		sh		$2, 2 + 1*16($25)
		dsrl	$2, $2, 16
		sw		$2, 4 + 1*16($25)
		sd		$10, 8 + 1*16($25)
		pcpyud	$2, $10, $0
		sh		$2, 14 + 1*16($25)

		# write third quad
		dsrl	$2, $2, 16
		sw		$2, 0 + 2*16($25)
		sw		$11, 4 + 2*16($25)
		pcpyud	$2, $11, $0
		dsrl	$11, $11, 32
		sh		$11, 8 + 2*16($25)
		sh		$2, 10 + 2*16($25)
		dsrl	$2, $2, 16
		sw		$2, 12 + 2*16($25)
		addiu	$24, $24, 64

		bne		$24, $3, 0b
		addiu	$25, $25, 48

		.set	pop		"	: : "r" (pColorsOut), "r" (pColorsIn), "r" (iNumCacheLinesIn) :	"$2", "$3", "$8", "$9", "$10", "$11", "$14", "$15", "$24", "$25");

}


// cacheline is 64bytes (16 colors in rgba)
// This is heavily optimized for when
// scratchpad is both source and destination.
void _gsvncConvRGBA_2(void * pColorsOut, const void * pColorsIn, const sps2uint32 iNumCacheLinesIn) {
	// no pref+no ucab:		93 cycles per itteration (output: 16 x 24bit colors)
	// pref+no ucab:		93 cycles per itteration (output: 16 x 24bit colors)
	// pref+ucab:			88 cycles per itteration (output: 16 x 24bit colors)
	// scratchpad:			27 cycles per itteration (output: 16 x 24bit colors)

	asm __volatile__ ( "
		.set	push
		.set	mips3
		
		sll		$3, %2, 6		# mul by 64
		move	$24, %1			# src colors
		add		$3, %1, $3		# stop_pointer (offset by source pointer)
		move	$25, %0			# dst colors
		li		$8, 0x00ffffff
		pcpyld	$14, $8, $8
		dsll	$8, $8, 24
		pcpyld	$15, $8, $8
		mtsab	$0, 1

		addiu	$25, $25, -48



		# epilogue
		lq		$8, 0($24)
		qfsrv	$2, $0, $8
		lq		$16, 16($24)
		pand	$8, $14, $8
		lq		$17, 32($24)
		pand	$2, $15, $2
		lq		$18, 48($24)
		por		$8, $2, $8
		qfsrv	$2, $0, $16
		pand	$9, $14, $16
		pand	$2, $15, $2
		por		$9, $2, $9
		qfsrv	$2, $0, $17
		pand	$10, $14, $17
		pand	$2, $15, $2
		por		$10, $2, $10
		qfsrv	$2, $0, $18
		pand	$11, $14, $18
		pand	$2, $15, $2
		por		$11, $2, $11
		sd		$8, 0 + 0*16+48($25)
		pcpyud	$2, $8, $0
		dsrl	$4, $2, 16
		dsrl	$5, $9, 32
		sh		$2, 6 + 0*16+48($25)
		pcpyud	$2, $9, $0
		sh		$2, 2 + 1*16+48($25)
		pcpyud	$13, $10, $0
		dsrl	$6, $2, 16
		dsrl	$7, $13, 16
		sd		$10, 8 + 1*16+48($25)
		pcpyud	$2, $11, $0
		dsrl	$12, $11, 32
		dsrl	$10, $2, 16
		sh		$2, 10 + 2*16+48($25)
		addiu	$24, $24, 64
		addiu	$25, $25, 48
		beq		$24, $3, 1f
		lq		$8, 0($24)


		.set	noreorder
		.align	4

		
		# loop	, 26 dual issued (paired instructions)
	0:
		mfc1	$0, $f0
		qfsrv	$2, $0, $8
		
		lq		$16, 16($24)
		pand	$8, $14, $8
		
		lq		$17, 32($24)
		pand	$2, $15, $2
		
		lq		$18, 48($24)
		por		$8, $2, $8

		
		sw		$9, 12 + 0*16($25)
		qfsrv	$2, $0, $16

		sw		$10, 12 + 2*16($25)
		pand	$9, $14, $16
		
		sw		$11, 4 + 2*16($25)
		pand	$2, $15, $2

		sw		$4, 8 + 0*16($25)
		por		$9, $2, $9

		sh		$5, 0 + 1*16($25)
		qfsrv	$2, $0, $17

		sw		$6, 4 + 1*16($25)
		pand	$10, $14, $17


		sw		$7, 0 + 2*16($25)
		pand	$2, $15, $2

		sh		$12, 8 + 2*16($25)
		por		$10, $2, $10

		sh		$13, 14 + 1*16($25)
		qfsrv	$2, $0, $18
		
		mfc1	$0, $f0
		pand	$11, $14, $18

		mfc1	$0, $f0
		pand	$2, $15, $2

		mfc1	$0, $f0
		por		$11, $2, $11

		sd		$8, 0 + 0*16+48($25)
		pcpyud	$2, $8, $0

		dsrl	$4, $2, 16
		dsrl	$5, $9, 32

		sh		$2, 6 + 0*16+48($25)
		pcpyud	$2, $9, $0
		

		sh		$2, 2 + 1*16+48($25)
		pcpyud	$13, $10, $0

		dsrl	$6, $2, 16
		dsrl	$7, $13, 16


		sd		$10, 8 + 1*16+48($25)
		pcpyud	$2, $11, $0

		dsrl	$12, $11, 32
		dsrl	$10, $2, 16
		
		sh		$2, 10 + 2*16+48($25)
		addiu	$24, $24, 64

		addiu	$25, $25, 48
		lq		$8, 0($24)

		bne		$24, $3, 0b
                pref            0, 48*2($25)
		# nop

	1:
		# prologue
		sw		$9, 12 + 0*16($25)
		sw		$10, 12 + 2*16($25)
		sw		$11, 4 + 2*16($25)
		sw		$4, 8 + 0*16($25)
		sh		$5, 0 + 1*16($25)
		sw		$6, 4 + 1*16($25)
		sw		$7, 0 + 2*16($25)
		sh		$12, 8 + 2*16($25)
		sh		$13, 14 + 1*16($25)
		

		.set	pop		"	: : "r" (pColorsOut), "r" (pColorsIn), "r" (iNumCacheLinesIn) :	"$2", "$3", "$4", "$5", "$6", "$7", "$8", "$9", "$10", "$11", "$12", "$13", "$14", "$15", "$24", "$25");
}
