/**
 * Please note that this code may not be specifically used by any of the
 * samples; however, it is provided as a template for a screen class should
 * you need one that provides more functionality than the basic screen.cpp/h.
 * The basic idea is that you would copy screen_template.cpp/h over to your
 * own source tree, rename it appropriately, and modify it to add whatever
 * functionality you would need.
 */
#ifndef __SCREEN_H__
#define __SCREEN_H__

#include "exttypes.h"

#define SCR_VESA		0
#define SCR_NTSC		2
#define SCR_PAL		3
#define SCR_WIDTH		640
#define SCR_HEIGHT_VESA	480
#define SCR_HEIGHT_NTSC	448
#define SCR_HEIGHT_PAL	512
#define SCR_DEPTH		0xffffff		// 24bit


class CScreen
{
	friend void  DistribGlobPtrs(void * pDescCached, void * pDescUncached);
	
public:
	void ScreenSetup();
	void Swap();
	uint32 GetFirstFreeGsPage();
	uint32 GetDrawFptr();
	uint32 GetZptr();
	uint32 GetDrawBuff1();
	uint32 GetDrawBuff2();
	uint32 GetWidth() { return m_uWidth; }
	uint32 GetHeight() { return m_uHeight; }
	uint32 GetDepth() { return m_uDepth; }
	uint32 GetMode() { return m_uMode; }

	CScreen();
	~CScreen();

private:
	// PS2Linux stuff
	void reset_gs_vesa();
	void reset_gs_ntsc();
	void reset_gs_pal();
	void reset_gs(int inter, int omode, int ffmode, int resolution, int refreshRate);
	void VSync(int mode);
	void virtual_console_start();
	void virtual_console_stop();
	uint32 auto_detect_video_mode();

	int32	m_fd_gs, m_fd_event, m_inter;
	void * m_pOldScrInfo;

	int m_console_fd;
	int m_tty_fd;
	int m_startup_vc;

private:
	void SetBuffers();
	void Setup();

	uint32 m_uWidth, m_uHeight, m_uDepth, m_uMode;
	uint32 buffer;

	uint32 m_uPtr1, m_uPtr2;			// Draw + Disp pointer (in pages)
	uint32 m_zBufPtr;					// ZBuf pointer (in pages)
	uint32 m_uRemBufPtr;				// pointer to reamaining buffer space (in pages)
};


#endif
