/*
    Copyright (C) 2002 Terratron Technologies Inc.  All Rights Reserved.
    Author: Steven Osman, Morten Mikkelsen, Lionel Lemarie
 
    This file is part of sps2.

    Before using this file you MUST agree to the license agreement in
    the file LICENSE provided with this package.
  
    For more information visit:
    http://www.playstation2-linux.com/projects/sps2/
    http://window.terratron.com/~sosman/ps2linux/
    http://www.terratron.com/

    7/12/2003 - Sauce
    Initial version
*/
#include <sys/ioctl.h>
#include <sys/vt.h>
#include <linux/ps2/dev.h>
#include <linux/kd.h>
#include <stdio.h>
#include <assert.h>
#include <signal.h>
#include <stdarg.h>

#include "sps2lib.h"
#include "sps2util.h"
#include "gsvnc_sps2.h"
#include "sps2tags.h"
#include "sps2utilfontfixed.h"

/**
 * These variables are common to the sps2U library
 */
void *_pvSPS2UEERegisters;
void *_pvSPS2UGSRegisters;
int _iSPS2USPS2Descriptor=-1;

/**
 * These functions are common to the sps2U library
 */
void _sps2UOpenSPS2();

/*****************************************************************************
 * SPS2U Common Private Functions
 *****************************************************************************/
/**
 * This function opens the copy of SPS2 that will be used by all of the
 * sps2 utility functions.
 * It will also set a common set of functionality (e.g. mapping EE & GS
 * registers).
 */
void _sps2UOpenSPS2() {
  if (_iSPS2USPS2Descriptor<0) {
    _iSPS2USPS2Descriptor=_sps2Open();
    
    _pvSPS2UEERegisters=_sps2MapEERegisters(0,_iSPS2USPS2Descriptor);
    _pvSPS2UGSRegisters=_sps2MapGSRegisters(0,_iSPS2USPS2Descriptor);
  }
}

/**
 * These macros are used by sps2UScreen
 */
#define DISPFB_SET(FBP, FBW, PSM, DBX, DBY)					       (((sps2uint64) DBY<<43) | ((sps2uint64) DBX<<32) |			       ((sps2uint64) PSM<<15) | ((sps2uint64) FBW<<9) |				      ((sps2uint64) FBP<<0))

#define GIFTAG(NLOOP, EOP, PRE, PRIM, FLG, NREG)				       ((((sps2uint64)NREG)<<60) | (((sps2uint64)FLG)<<58) |			       (((sps2uint64)PRIM)<<47) | (((sps2uint64)PRE)<<46) |			      (EOP<<15) | (NLOOP<<0))

#define GIFTAG_REGS16(a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p)							((((sps2uint64)p)<<60)|(((sps2uint64)o)<<56)|(((sps2uint64)n)<<52)|(((sps2uint64)m)<<48)|(((sps2uint64)l)<<44)|(((sps2uint64)k)<<40)|(((sps2uint64)j)<<36)|(((sps2uint64)i)<<32)|(((sps2uint64)h)<<28)|(((sps2uint64)g)<<24)|(((sps2uint64)f)<<20)|(((sps2uint64)e)<<16)|(((sps2uint64)d)<<12)|(((sps2uint64)c)<<8)|(((sps2uint64)b)<<4)|(((sps2uint64)a)<<0))
#define GIFTAG_REGS12(a,b,c,d,e,f,g,h,i,j,k,l)									GIFTAG_REGS16(a,b,c,d,e,f,g,h,i,j,k,l,0,0,0,0)
#define GIFTAG_REGS8(a,b,c,d,e,f,g,h)											GIFTAG_REGS12(a,b,c,d,e,f,g,h,0,0,0,0)
#define GIFTAG_REGS4(a,b,c,d)													GIFTAG_REGS8(a,b,c,d,0,0,0,0)
#define GIFTAG_REGS1(a)															GIFTAG_REGS4(a,0,0,0)

#define PRMODE																	((sps2uint64) 0x1b)
#define PRMODE_SET(IIP, TME, FGE, ABE, AA1, FST, CTXT, FIX)						((sps2uint64) ((FIX<<10)|(CTXT<<9)|(FST<<8)|(AA1<<7)|(ABE<<6)|(FGE<<5)|(TME<<4)|(IIP<<3)))

#define PRIM_REG_NO																	((sps2uint64) 0x00)
#define PRIM_SET(PRIM, IIP, TME, FGE, ABE, AA1, FST, CTXT, FIX)					(PRMODE_SET(IIP, TME, FGE, ABE, AA1, FST, CTXT, FIX)|PRIM)

#define AD																		((sps2uint64) 0x0e)
#define FRAME_1																	((sps2uint64) 0x4c)
#define FRAME_SET(FBP, FBW, PSM, FBMSK)											((((sps2uint64)FBMSK)<<32) | (((sps2uint64)PSM)<<24) | (((sps2uint64)FBW)<<16) | ((sps2uint64)FBP))


#define ZBUF_1																	((sps2uint64) 0x4e)
#define ZBUF_SET(ZBP, PSM, ZMSK)												((((sps2uint64)ZMSK)<<32) | (((sps2uint64)PSM)<<24) | (((sps2uint64)ZBP)<<0))

#define XYOFFSET_1																((sps2uint64) 0x18)
#define XYOFFSET_SET(OFX, OFY)													((((sps2uint64)OFY)<<32) | (((sps2uint64)OFX)<<0))

#define SCISSOR_1																((sps2uint64) 0x40)
#define SCISSOR_SET(SCAX0, SCAX1, SCAY0, SCAY1)									((((sps2uint64)SCAY1)<<48) | (((sps2uint64)SCAY0)<<32) | (((sps2uint64)SCAX1)<<16) | (((sps2uint64)SCAX0)<<0))

#define COLCLAMP																((sps2uint64) 0x46)
#define COLCLAMP_SET(CLAMP)														((sps2uint64) CLAMP&0x1)

#define DTHE																	((sps2uint64) 0x45)
#define DTHE_SET(DTHE)															((sps2uint64) DTHE&0x1)

#define TEST_1																	((sps2uint64) 0x47)
#define TEST_SET(ATE, ATST, AREF, AFAIL, DATE, DATM, ZTE, ZTST)					((((sps2uint64)ZTST)<<17) | (((sps2uint64)ZTE)<<16) | (((sps2uint64)DATM)<<15) | (((sps2uint64)DATE)<<14) | (((sps2uint64)AFAIL)<<12) | (((sps2uint64)AREF)<<4) | (((sps2uint64)ATST)<<1) | (((sps2uint64)ATE)<<0))

#define RGBAQ_REG_NO																	((sps2uint64) 0x01)
#define RGBAQ_SET(R,G,B,A,Q)													((((sps2uint64)Q)<<32) | (((sps2uint64)A)<<24) | (((sps2uint64)B)<<16) | (((sps2uint64)G)<<8) | (((sps2uint64)R)<<0))

#define XYZ2_REG_NO																	((sps2uint64) 0x05)
#define XYZ2_SET(X, Y, Z)														((((sps2uint64)Z)<<32) | (((sps2uint64)Y)<<16) | (((sps2uint64)X)<<0))

#define PRMODECONT																((sps2uint64) 0x1a)
#define PRMODECONT_SET(AC)														((sps2uint64) AC&0x1)


#define SPS2U_SCR_WIDTH		640
#define SPS2U_SCR_HEIGHT_VESA	480
#define SPS2U_SCR_HEIGHT_NTSC	448
#define SPS2U_SCR_HEIGHT_PAL	512
#define SPS2U_SCR_DEPTH		0xffffff		// 24bit
#define SPS2_SCR_PSMCT32        0

/**
 * These variables are used by sps2UScreen
 */
int _iSPS2UUseGSVNC=1;
int _iSPS2UGSVNCUpdateRate=15;
sps2int32 mscr_fd_gs, mscr_fd_event, mscr_inter;
void * mscr_pOldScrInfo;

int mscr_console_fd;
int mscr_tty_fd;
int mscr_startup_vc;

sps2uint32 mscr_uWidth, mscr_uHeight, mscr_uDepth, mscr_uVideoMode;
sps2uint32 mscr_uPixelFormat=SPS2_SCR_PSMCT32;
sps2uint32 mscr_iCurrentDrawBuffer;

// Draw + Disp pointer (in pages)
sps2uint32 mscr_uBufferPointer1, mscr_uBufferPointer2;
// ZBuf pointer (in pages)
sps2uint32 mscr_zBufPtr;
// pointer to reamaining buffer space (in pages)
sps2uint32 mscr_uRemBufPtr;

// This is set to non-zero if we've hooked our signal handlers.
int mscr_SignalHandlers=0;

//
typedef void (*sps2UScreenSignalHandler_t)(int);
sps2UScreenSignalHandler_t mscr_OldSignalHandlers[128];

/**
 * These are internal functions needed by sps2UScreen
 */
void _sps2UScreenDoSwap();
void _sps2UScreenSignalHandler(int iSignal);
void _sps2UScreenSetBuffers();
void _sps2UScreenSetup();
void _sps2UScreenResetGSVESA();
void _sps2UScreenResetGSNTSC();
void _sps2UScreenResetGSPAL();
void _sps2UScreenResetGS(int inter, int omode, int ffmode,
		   int resolution, int refreshRate);
void _sps2UScreenVSync(int mode);
void _sps2UScreenVirtualConsoleStart();
void _sps2UScreenVirtualConsoleStop();
sps2uint32 _sps2UScreenAutoDetectVideoMode();

/*****************************************************************************
 * SPS2UScreen Public Functions
 *****************************************************************************/
void sps2UScreenInit(int iSuppressSignalHandlers) {
  int iSignalLooper;

  mscr_SignalHandlers=!iSuppressSignalHandlers;
  _sps2UScreenVirtualConsoleStart();
  
  if (mscr_SignalHandlers) {
    // Catch all stignals we can so that if there is a crash we can attempt
    // to shut down gracefully
    for(iSignalLooper=0; iSignalLooper<128; iSignalLooper++) {
      // SIGSTOP (i.e. control+Z) is harmless
      // So is SIGCONT
      if ((iSignalLooper==SIGSTOP) ||
	  (iSignalLooper==SIGCONT)) {
	continue;
      }

      mscr_OldSignalHandlers[iSignalLooper]=
	signal(iSignalLooper,_sps2UScreenSignalHandler);
    }
  }

  atexit(sps2UScreenShutdown);

  mscr_pOldScrInfo = (void *) malloc(sizeof(struct ps2_screeninfo));
	
  mscr_iCurrentDrawBuffer = 0;
  mscr_inter = PS2_GS_NOINTERLACE;
	
  mscr_fd_gs = open(PS2_DEV_GS, O_RDWR);
  if(mscr_fd_gs < 0)
    {
      perror("open(PS2_DEV_GS)");
      exit(0);
    }
  mscr_fd_event = open(PS2_DEV_EVENT, O_RDWR);
  if(mscr_fd_event < 0)
    {
      perror("open(PS2_DEV_EVENT)");
      close(mscr_fd_gs);
      exit(0);
    }

  ioctl(mscr_fd_gs, PS2IOC_GSCREENINFO, 
	(struct ps2_screeninfo *) mscr_pOldScrInfo);
  ioctl(mscr_fd_event, PS2IOC_ENABLEEVENT, PS2EV_VBSTART);

  _sps2UScreenSetup();

  sps2UPrintfSetWindow(-(int) sps2UScreenGetWidth()/2, 
		       -(int) sps2UScreenGetHeight()/2,
		       sps2UScreenGetWidth(), sps2UScreenGetHeight());

  if (_iSPS2UUseGSVNC) {
    _iSPS2UUseGSVNC=!gsvncSPS2Init(1);
  }
}

void sps2UScreenShutdown()
{
  int iSignalLooper;

  if (mscr_SignalHandlers) {
    // Catch all stignals we can so that if there is a crash we can attempt
    // to shut down gracefully
    for(iSignalLooper=0; iSignalLooper<128; iSignalLooper++) {
      // Ignore signals that hadn't been trapped
      if ((iSignalLooper==SIGSTOP) ||
	  (iSignalLooper==SIGCONT)) {
	continue;
      }

      signal(iSignalLooper, mscr_OldSignalHandlers[iSignalLooper]);
    }

    mscr_SignalHandlers=0;
  }

  if (mscr_fd_gs==-1) return;
  ioctl(mscr_fd_gs, PS2IOC_SSCREENINFO,
	(struct ps2_screeninfo *) mscr_pOldScrInfo);
  if(mscr_fd_event >= 0) close(mscr_fd_event);
  if(mscr_fd_gs >= 0) close(mscr_fd_gs);
  free ((struct ps2_screeninfo *) mscr_pOldScrInfo);
  _sps2UScreenVirtualConsoleStop();
  mscr_fd_gs=-1;
}

void sps2UScreenSwap() {
  if (_iSPS2UUseGSVNC) {
    gsvncSPS2UtilUpdate(_iSPS2UGSVNCUpdateRate);
  }

  _sps2UScreenDoSwap();
}

void _sps2UScreenDoSwap() {
  sps2uint32 uNewBuffer=mscr_iCurrentDrawBuffer=1-mscr_iCurrentDrawBuffer;
  sps2uint32 uPage=(uNewBuffer == 0) ?
    mscr_uBufferPointer1 : mscr_uBufferPointer2;
  sps2uint32 uDispPage=(uNewBuffer == 0) ?
    mscr_uBufferPointer2 : mscr_uBufferPointer1;

  sps2uint32 uMask = 0;		
  //  sps2uint32 uMode = SPS2_SCR_PSMCT32;	// 32bit
  sps2uint32 q;
  // various data args
  const sps2uint32 nloop = 7;
  const sps2uint32 uOffsX = (2048-(mscr_uWidth>>1))*16;
  // Now center in render will always be 2048, 2048
  const sps2uint32 uOffsY = (2048-(mscr_uHeight>>1))*16;

  // gif codes data
  const sps2uint64 uGifData[] = // _ALIGN16 =
  {
    GIFTAG(nloop, 1, 0, 0, 0, 1),
    GIFTAG_REGS1(AD),

    FRAME_SET(uPage,mscr_uWidth/64,mscr_uPixelFormat,uMask),
    FRAME_1,

    ZBUF_SET(mscr_zBufPtr, 1, 0),
    ZBUF_1,

    XYOFFSET_SET(uOffsX, uOffsY),
    XYOFFSET_1,

    SCISSOR_SET(0, mscr_uWidth-1, 0, mscr_uHeight-1),
    SCISSOR_1,

    PRMODECONT_SET(0x1),
    PRMODECONT,

    COLCLAMP_SET(0x1),
    COLCLAMP,
		
    DTHE_SET(0x0),
    DTHE,
  };

  // Wait for GIF to free up
  while (*EE_GIF_STAT_OFF(_pvSPS2UEERegisters) & (31 << 24)); 
  _sps2UScreenVSync(0);	       // Wait for vertical blank

  DPUT_GS_DISPFB2_OFF(_pvSPS2UGSRegisters, 
		      DISPFB_SET(uDispPage, mscr_uWidth/64, 
				 mscr_uPixelFormat, 0, 0));

  // execute gif codes
  for(q=0; q<nloop+1; q++)
    DPUT_EE_GIF_FIFO_OFF(_pvSPS2UEERegisters, ((sps2uint128 *) uGifData)[q]);

  //GsSyncPath(0, 0);
}

void sps2UScreenClear(int iRed, int iGreen, int iBlue) {
  const sps2uint32 nloop = 6;
  const sps2uint32 uOffsX = (2048-(mscr_uWidth>>1))*16;
  // Now center in render will always be 2048, 2048
  const sps2uint32 uOffsY = (2048-(mscr_uHeight>>1))*16;
  const sps2uint32 uEndX = (2048+(mscr_uWidth>>1))*16;
  const sps2uint32 uEndY = (2048+(mscr_uHeight>>1))*16;

  int q;

  // gif codes data
  const sps2uint64 uGifData[] = // _ALIGN16 =
  {
    GIFTAG(nloop, 1, 0, 0, 0, 1),
    GIFTAG_REGS1(AD),

    ///////////////////////// This last part is added to clear the screen
    TEST_SET(0, 0, 0, 0, 0, 0, 1, 1),
    TEST_1,

    PRIM_SET(0x6, 0, 0, 0, 0, 0, 0, 0, 0),
    PRIM_REG_NO,
		
    RGBAQ_SET(iRed, iGreen, iBlue, 0x80, 0x3f800000),
    RGBAQ_REG_NO,

    XYZ2_SET(uOffsX, uOffsY, 0),
    XYZ2_REG_NO,

    XYZ2_SET(uEndX, uEndY, 0),
    XYZ2_REG_NO,
	
    TEST_SET(0, 0, 0, 0, 0, 0, 1, 3),
    TEST_1
  };

  // execute gif codes
  for(q=0; q<nloop+1; q++)
    DPUT_EE_GIF_FIFO_OFF(_pvSPS2UEERegisters, ((sps2uint128 *) uGifData)[q]);
}
  
unsigned int sps2UScreenGetFirstFreeGSPage()
{
  return mscr_uRemBufPtr;
}
	
unsigned int sps2UScreenGetZPtr()
{
  return mscr_zBufPtr;
}

unsigned int sps2UScreenGetDrawBuff1()
{
  return mscr_uBufferPointer1;
}

unsigned int sps2UScreenGetDrawBuff2()
{
  return mscr_uBufferPointer2;
}

unsigned int sps2UScreenGetDrawBuffCurrent()
{
  return (mscr_iCurrentDrawBuffer == 0) ? mscr_uBufferPointer1 : 
    mscr_uBufferPointer2;
}

unsigned int sps2UScreenGetWidth() 
{
  return mscr_uWidth; 
}

unsigned int sps2UScreenGetHeight()
{
  return mscr_uHeight;
}

unsigned int sps2UScreenGetZDepth()
{
  return mscr_uDepth;
}

unsigned int sps2UScreenGetVideoMode()
{
  return mscr_uVideoMode;
}

unsigned int sps2UScreenGetPixelFormat()
{
  return mscr_uPixelFormat;
}

  /**
   * Called _before_ sps2UScreenInit, this disables the use of gsvnc.
   * Note that gsvnc offers only a small memory overhead and no performance
   * overhead if you don't have a vnc server running _and_ a client connected
   * to it, so generally you'd want to leave the support in there.
   */
void sps2UScreenDisableVNC() {
  _iSPS2UUseGSVNC=0;
}

/**
 * Sets the frequency at which the vnc server should get an updated frame.
 * Note that frame grabs only happen if there's a client connected to the
 * vnc server.
 */
void sps2UScreenSetVNCUpdateRate(int iUpdateRate) {
  _iSPS2UGSVNCUpdateRate=iUpdateRate;
}

/*****************************************************************************
 * SPS2UScreen Private Functions
 *****************************************************************************/

/**
 * This is our signal handler.  Since it doesn't really know what to do when
 * it gets a signal it will just abort the program.  This beats the program
 * crashing when it gets the unknown signal because the display code will
 * have a chance to shut itself down.
 * iSignal is the signal being delivered
 */
void _sps2UScreenSignalHandler(int iSignal) {
  printf("sps2UScreen: Aborting  on signal %i\n", iSignal);
  exit(0);
}

void _sps2UScreenSetup()
{
  //sps2uint32 uVCKv = 1;
  sps2uint32 uGSPage = 0;
  //sps2uint32 uMask = 0;		
  //  sps2uint32 uMode = SPS2_SCR_PSMCT32;	// 32bit
  //sps2uint32 uVCKh, uDispX, uDispY;

  _sps2UOpenSPS2();

  //GsResetPath();
  //DmaReset(1);
  *EE_VIF0_ERR_OFF(_pvSPS2UEERegisters) = 2;
  *EE_VIF1_ERR_OFF(_pvSPS2UEERegisters) = 2;

// 512 pages in buffer

  // Setup Vesa
  switch(_sps2UScreenAutoDetectVideoMode())
    {
    case SPS2U_SCR_VESA:
      _sps2UScreenResetGSVESA();
      break;
    case SPS2U_SCR_NTSC:
      _sps2UScreenResetGSNTSC();
      break;
    case SPS2U_SCR_PAL:
      _sps2UScreenResetGSPAL();
      break;
    default:
      exit(0);
    }
	
  /*if(uScrMode == 0)	// NTSC	(640x448, 32bit, field, interlace)
	{
		uVCKh = 4;		// PixelSize in VCK units on screen 
		uDispX = 0x27c;		// display posX in VCK units, for some reason 
		uDispY = 0x32;
		mscr_uWidth = 640;
		mscr_uHeight = 448;
		//GsResetGraph(0, GS_INTERLACE, GS_NTSC, GS_FIELD);	// read each line
	}
	else			// PAL	(512x512, 32bit, field, interlace)
	{
		uVCKh = 5;
		uDispX = 0x290;
		uDispY = 0x48;
		mscr_uWidth = 512;
		mscr_uHeight = 512;
		//GsResetGraph(0, GS_INTERLACE, GS_PAL, GS_FIELD);	// read each line
	}*/

	//FlushCache(0);
  *EE_GIF_MODE_OFF(_pvSPS2UEERegisters) = 4;	// enable intermittent mode.

  // *GS_PMODE(mscr__pvSPS2UGSRegisters) = PMODE_SET(0, 1, 1, 1, 1, 0, 0);
  // *GS_SMODE2(mscr__pvSPS2UGSRegisters) = SMODE2_SET(0,0,0);
  DPUT_GS_DISPFB2_OFF(_pvSPS2UGSRegisters,
		      DISPFB_SET(uGSPage, mscr_uWidth/64, 
				 mscr_uPixelFormat, 0, 0));
  // *GS_DISPLAY2(mscr__pvSPS2UGSRegisters) = DISPLAY_SET(uDispX, uDispY, uVCKh-1, uVCKv-1, mscr_uWidth*uVCKh-1, mscr_uHeight-1);
  // *GS_BGCOLOR(mscr__pvSPS2UGSRegisters) = 0x0;

  _sps2UScreenSetBuffers();
}

void _sps2UScreenSetBuffers()
{
  // locked at 32bit
  // 512 pages in GS buffer
  mscr_uBufferPointer1=0;
  mscr_uBufferPointer2=(mscr_uWidth*mscr_uHeight)/2048;			
  mscr_zBufPtr=mscr_uBufferPointer2+((mscr_uWidth*mscr_uHeight)/2048);
  mscr_uRemBufPtr=mscr_zBufPtr+((mscr_uWidth*mscr_uHeight)/2048);
  assert(((mscr_uWidth*mscr_uHeight) % 2048) == 0);
}

void _sps2UScreenResetGSVESA()
{	
  mscr_uWidth = SPS2U_SCR_WIDTH;
  mscr_uHeight = SPS2U_SCR_HEIGHT_VESA;
  mscr_uDepth = SPS2U_SCR_DEPTH;
  _sps2UScreenResetGS(PS2_GS_NOINTERLACE, PS2_GS_VESA, 0, PS2_GS_640x480,
		      PS2_GS_60Hz);
}

void _sps2UScreenResetGSNTSC()
{	
  mscr_uWidth = SPS2U_SCR_WIDTH;
  mscr_uHeight = SPS2U_SCR_HEIGHT_NTSC;
  mscr_uDepth = SPS2U_SCR_DEPTH;
  _sps2UScreenResetGS(PS2_GS_INTERLACE, PS2_GS_NTSC, PS2_GS_FIELD, 
		      0/*PS2_GS_640x448*/, 0);
}

void _sps2UScreenResetGSPAL()
{	
  mscr_uWidth = SPS2U_SCR_WIDTH;
  mscr_uHeight = SPS2U_SCR_HEIGHT_PAL;
  mscr_uDepth = SPS2U_SCR_DEPTH;
  _sps2UScreenResetGS(PS2_GS_INTERLACE, PS2_GS_PAL, PS2_GS_FIELD, 
		      0/*PS2_GS_640x512*/, 0);
}

void _sps2UScreenResetGS(int inter, int omode, int ffmode, int resolution,
			 int refreshRate)
{
  //ps2_gs_reset(0, PS2_GS_NOINTERLACE, PS2_GS_VESA, PS2_GS_FRAME,
  // PS2_GS_640x480, PS2_GS_60Hz);
  // args
	
  // setup code
  int ret;
  __u64 csr;
		
	//struct ps2_screeninfo sinfo;
  struct ps2_crtmode crtmode;
  struct ps2_display display2;
  struct ps2_pmode pmode;
  int width, height;
  int interlaceFormat;

  int reg = PS2_GSSREG_CSR;
  struct ps2_gssreg r;
	
  mscr_inter = inter;

  // full reset GS
  /*ret = ioctl(mscr_fd_gs, PS2IOC_GSRESET, PS2_GSRESET_FULL);
    if (ret < 0) {
    //perror("ps2_gs_reset, ioctl PS2_GSRESET_FULL");
    exit(0);
    }*/
		
  // set pmode, display off
  ret = ioctl(mscr_fd_gs, PS2IOC_GPMODE, &pmode);
  pmode.sw = 0; // ch1: off, ch2: off
  ret = ioctl(mscr_fd_gs, PS2IOC_SPMODE, &pmode);
		
  interlaceFormat = (omode == PS2_GS_NTSC || omode == PS2_GS_PAL ||
		     (omode == PS2_GS_DTV && resolution == PS2_GS_1080I));

  // get GS version
  r.reg = reg;
  ioctl(mscr_fd_gs, PS2IOC_GGSSREG, &r/*eg*/);
  csr = r.val;
  //csr = __ps2_gs_get_sreg(PS2_GSSREG_CSR);
	
	
  width = 640;
  height = 480;
		
  switch (omode)
    {
    case PS2_GS_NTSC:
    case PS2_GS_PAL:
      width = 640;
      height = (omode == PS2_GS_NTSC) ? 448 : 512;
      break;
			
    case PS2_GS_DTV:
      switch (resolution) {
      case PS2_GS_480P: width = 720; height = 480; break;
      case PS2_GS_1080I: width = 1920; height = 1080; break;
      case PS2_GS_720P: width = 1024; height = 720; break;
      }
      break;
			
    case PS2_GS_VESA:
      switch (resolution) {
      case PS2_GS_640x480: width = 640; height = 480; break;
      case PS2_GS_800x600: width = 800; height = 600; break;
      case PS2_GS_1024x768: width = 1024; height = 768; break;
      case PS2_GS_1280x1024: width = 1280; height = 1024; break;
      }
    }
		
  if (interlaceFormat && 
      (mscr_inter == PS2_GS_NOINTERLACE || ffmode == PS2_GS_FRAME)) {
    // fake NI or I(frame)
    height /= 2;
  }
		
  //ps2_gs_set_center(2048, 2048);
  //__ps2_gs_set_screen_size(width, height);
	
  crtmode.mode = omode;
		
  switch (omode)
    {
    case PS2_GS_NTSC:
    case PS2_GS_PAL:
    case PS2_GS_DTV:
      crtmode.res = mscr_inter | ffmode;
      break;
			
    case PS2_GS_VESA:
      crtmode.res = resolution | refreshRate;
      break;
			
    default:
      //return -1;
      exit(0);
    }
		
  //ret = ioctl(gp->fd_gs, PS2IOC_SSCREENINFO, &sinfo);
  ret = ioctl(mscr_fd_gs, PS2IOC_SCRTMODE, &crtmode);
  if (ret < 0) {
    //perror("ps2_gs_reset, PS2IOC_SSCREENINFO");
    exit(0);
  }
		
  // display
  display2.ch = 1; // use region 2
  display2.w = width;
  display2.h = height;
  display2.dx = 0;
  display2.dy = 0;

  if (interlaceFormat && 
      (mscr_inter == PS2_GS_INTERLACE && ffmode == PS2_GS_FRAME))
    display2.h *= 2;
		
  ret = ioctl(mscr_fd_gs, PS2IOC_SDISPLAY, &display2);
  if (ret < 0) {
    //perror("ps2_gs_reset, PS2IOC_SDISPLAY");
    exit(0);
  }

  // set pmode, display on
  ret = ioctl(mscr_fd_gs, PS2IOC_GPMODE, &pmode);
  pmode.sw = 2; // ch1: off, ch2: off
  ret = ioctl(mscr_fd_gs, PS2IOC_SPMODE, &pmode);
}

void _sps2UScreenVSync(int mode)
{
  //ps2_gs_sync_v(mode);
  struct ps2_gssreg reg;
	
  ioctl(mscr_fd_event, PS2IOC_GETEVENT, PS2EV_VBSTART);
  ioctl(mscr_fd_event, PS2IOC_WAITEVENT, PS2EV_VBSTART);

	//int interlace_mode = 1;
	
  if (mscr_inter == PS2_GS_INTERLACE) {
    reg.reg = PS2_GSSREG_CSR;
    if (ioctl(mscr_fd_gs, PS2IOC_GGSSREG, &reg) < 0) {
      //perror("ps2_gs_sync_v, ioctl");
      exit(0);
    }
		
    // int iRetVal = ((ps2_gssreg_csr *)&reg.val)->FIELD;
  }
	
  // int iRetVal = 1; // odd
}

void _sps2UScreenVirtualConsoleStart()
{
  char fname[32];
  int mode;
  int new_vc;
  int err;
  struct vt_stat vts;

  mscr_console_fd = open( "/dev/console", O_RDWR );

  if (mscr_console_fd<0)
    {
      perror("open /dev/console");
      exit(-1);
    }

  err=ioctl(mscr_console_fd, VT_OPENQRY, &new_vc);
  if (err==-1)
    {
      perror("ioctl(VT_OPENQRY)");
      exit(-1);
    }
  if (new_vc<=0)
    {
      perror("ioctl(VT_OPENQRY) (wrong new_vc returned)");
      exit(-1);
    }

  sprintf(fname, "/dev/tty%d", new_vc);

  setpgid(0,getppid());
  setsid();

  mscr_tty_fd = open(fname, O_RDWR);
  if (mscr_tty_fd<0)
    {
      perror(fname);
      exit(-1);
    }

  ioctl(mscr_tty_fd, VT_GETSTATE, &vts);
  mscr_startup_vc=vts.v_active;

  ioctl(mscr_tty_fd, VT_ACTIVATE, new_vc);

  while (ioctl(mscr_tty_fd, VT_WAITACTIVE, new_vc) < 0)
    {
      if ((errno != EAGAIN) && (errno != EINTR))
	{
	  perror("ioctl(VT_WAITACTIVE)");
	  exit(-1);
	}
      usleep(150000);
    }


  // clear the console
  write(mscr_tty_fd, "\e[H\e[J", 6);


  mode = KD_GRAPHICS;
  printf("Switching to graphics\n");
  ioctl( mscr_tty_fd, KDSETMODE, mode );
}

void _sps2UScreenVirtualConsoleStop()
{
  int mode;

  mode = KD_TEXT;
  printf("Deleting new console\n");
  ioctl( mscr_tty_fd, KDSETMODE, mode );

  ioctl(mscr_tty_fd, VT_ACTIVATE, mscr_startup_vc);

  close( mscr_tty_fd );
  close( mscr_console_fd );
}

sps2uint32 _sps2UScreenAutoDetectVideoMode()
{
  int err;
  struct ps2_screeninfo screeninfo;

  err=ioctl(mscr_fd_gs, PS2IOC_GSCREENINFO, &screeninfo);
  if (err==-1)
    {
      perror("ioctl(PS2IOC_GSCREENINFO)");
      exit(-1);
    }

  switch (screeninfo.mode)
    {
    case 0:
      printf("Autodetected VESA mode\n");
      return mscr_uVideoMode=SPS2U_SCR_VESA;
    case 2:
      printf("Autodetected NTSC mode\n");
      return mscr_uVideoMode=SPS2U_SCR_NTSC;
    case 3:
      printf("Autodetected PAL mode\n");
      return mscr_uVideoMode=SPS2U_SCR_PAL;
    }

  fprintf(stderr, "Could not autodetect display mode. Using default (VESA)\n");
  return mscr_uVideoMode=SPS2U_SCR_VESA;
}


/*****************************************************************************
 * SPS2 Printf private structures & variables
 *****************************************************************************/

/**
 * This structure contains all the data we
 * will keep track of for each character being
 * displayed.  This could probably be optimized well
 * to compress data for characters that share the same
 * font and/or color.
 */
typedef struct {
  short sX;
  short sY;
  int iZ;
  sps2uint128 color;
  sps2UFontStruct *pFont;
  char cCharacter;
} sps2UPrintfEntry;

/**
 * The maximum number of characters we will hold as well as the buffer
 * to store them in.
 */
#define SPS2UPRINTF_BUFFER_SIZE 8192
sps2UPrintfEntry _paSPS2UPBuffer[SPS2UPRINTF_BUFFER_SIZE];

/**
 * The number of entries in the buffer
 */
int _iSPS2UPBufferCount=0;

/**
 * Point to the next line, according to the font's size.
 */
void _sps2UPAdvanceLine(sps2UFontStruct *pFont);

/**
 * This function returns the width of the specific font for the given
 * character.
 */
int _sps2UPCharWidth(sps2UFontStruct *pFont, char cChar);

/**
 * Prepares the gif tag & commands for a given character.
 */
void _sps2UPOutputEntry(sps2UPrintfEntry *pEntry);

/**
 * "cursor's" x location.
 */
int _iSPS2UPCurrentX=0;

/**
 * "cursor's" y location.
 */
int _iSPS2UPCurrentY=0;

/**
 * Top-right corner of the window + window dimensions
 * These are default values that can be changed with
 * sps2UPrintfSetWindow
 */
int _iSPS2UPStartX=0, _iSPS2UPStartY=0, _iSPS2UPWidth=640, _iSPS2UPHeight=200;
int _iSPS2UPZValue=128;

/*****************************************************************************
 * SPS2 Printf public functions
 *****************************************************************************/

/**
 * Sets the window that the text is output to.  Default is (x,y) at the top
 * and (iWidth, iHeight) to span the whole screen.
 * Note that changes to sps2UPrintfSetWindow only apply to the next set
 * of printfs, not the ones already in the queue.
 */
void sps2UPrintfSetWindow(int iStartXPosition, int iStartYPosition,
		       int iWindowWidth, int iWindowHeight) {
  _iSPS2UPStartX=_iSPS2UPCurrentX=iStartXPosition;
  _iSPS2UPStartY=_iSPS2UPCurrentY=iStartYPosition;
  _iSPS2UPWidth=iWindowWidth;
  _iSPS2UPHeight=iWindowHeight;
}

/**
 * Returns the window that the text is output to.
 */
void sps2UPrintfGetWindow(int *piX, int *piY, int *piWidth, int *piHeight) {
  if (piX) *piX=_iSPS2UPStartX;
  if (piY) *piY=_iSPS2UPStartY;
  if (piWidth) *piWidth=_iSPS2UPWidth;
  if (piHeight) *piHeight=_iSPS2UPHeight;
}

/**
 * Sets the current (x,y) coordinates relative to the top of the printing
 * window.  This is the same as embedding \376(iX, iY) to the front of your
 * next string.
 */
void sps2UPrintfSetPos(int iX, int iY) {
  _iSPS2UPCurrentX=iX+_iSPS2UPStartX;
  _iSPS2UPCurrentY=iY+_iSPS2UPStartY;
}

/**
 * Returns the current (x,y) coordinates relative to the top of the printing
 * window.
 */
void sps2UPrintfGetPos(int *piX, int *piY) {
  if (piX) *piX=_iSPS2UPCurrentX-_iSPS2UPStartX;
  if (piY) *piY=_iSPS2UPCurrentY-_iSPS2UPStartY;
}

/**
 * Sets the Z value for printing.  Defaults to 128.
 * Note that changes to sps2UPrintfSetZ only apply to the next set
 * of printfs, not the ones already in the queue.
 */
void sps2UPrintfSetZ(int iZ) {
  _iSPS2UPZValue=iZ;
}

/**
 * Returns the Z value being used for the text.
 */
int sps2UPrintfGetZ() {
  return _iSPS2UPZValue;
}

/**
 * Renders the buffered text.
 */ 
void sps2UPrintfRender() {
  int iLooper;

  for (iLooper=0;iLooper<_iSPS2UPBufferCount;iLooper++) {
    _sps2UPOutputEntry(&_paSPS2UPBuffer[iLooper]);
  }

  _iSPS2UPCurrentX=_iSPS2UPStartX;
  _iSPS2UPCurrentY=_iSPS2UPStartY;
  _iSPS2UPBufferCount=0;
}

/**
 * Prints a string to the GS.
 */
int sps2UPrintf(sps2UFontStruct *pFont, const char *format, ...) {
  static char acStringBuffer[2048];
  va_list arguments;
  char *pcCurrent;
  int iReturn=0;
  int iRed=0xff, iGreen=0xff, iBlue=0xff;
  sps2GIFPackedRegister_t gifRegister;

  gifRegister.i128=0;
  gifRegister.RGBAQ.s.R=iRed;
  gifRegister.RGBAQ.s.G=iGreen;
  gifRegister.RGBAQ.s.B=iBlue;
  
  if (!pFont) {
    pFont=&_sps2UPFontFixed;
  }

  va_start(arguments, format);

  vsnprintf(acStringBuffer, sizeof(acStringBuffer), format, arguments);

  pcCurrent=acStringBuffer;

  while (*pcCurrent) {
    switch (*pcCurrent) {
    case '\n':
      _sps2UPAdvanceLine(pFont);
      iReturn++;
      break;
    case '\377':
      // Look for color and change it.
      {
	int iLooper;
	sscanf(pcCurrent+1,"%2X%2X%2X",&iRed,&iGreen,&iBlue);
	gifRegister.i128=0;
	gifRegister.RGBAQ.s.R=iRed;
	gifRegister.RGBAQ.s.G=iGreen;
	gifRegister.RGBAQ.s.B=iBlue;
	for (iLooper=0;iLooper<6 && *pcCurrent;iLooper++, pcCurrent++);
      }
      break;
    case '\376':
      // Look for x,y and set it.
      {
	sscanf(pcCurrent+1,"(%i,%i)", &_iSPS2UPCurrentX, &_iSPS2UPCurrentY);
	_iSPS2UPCurrentX+=_iSPS2UPStartX;
	_iSPS2UPCurrentY+=_iSPS2UPStartY;

	while (*pcCurrent && *pcCurrent!=')') {
	  pcCurrent++;
	}

	if (*pcCurrent==')')
	  pcCurrent++;
      }
    default:
      if (_iSPS2UPBufferCount<SPS2UPRINTF_BUFFER_SIZE) {
	int iCharWidth=_sps2UPCharWidth(pFont, *pcCurrent);

	if (iCharWidth>0) {
	  if (_iSPS2UPCurrentX+iCharWidth>_iSPS2UPStartX+_iSPS2UPWidth) {
	    _sps2UPAdvanceLine(pFont);
	  }
	  
	  if (_iSPS2UPCurrentY<_iSPS2UPStartY+_iSPS2UPHeight) {
	    iReturn++;

	    _paSPS2UPBuffer[_iSPS2UPBufferCount].color=gifRegister.i128;
	    _paSPS2UPBuffer[_iSPS2UPBufferCount].sX=_iSPS2UPCurrentX;
	    _paSPS2UPBuffer[_iSPS2UPBufferCount].sY=_iSPS2UPCurrentY;
	    _paSPS2UPBuffer[_iSPS2UPBufferCount].iZ=_iSPS2UPZValue;
	    _paSPS2UPBuffer[_iSPS2UPBufferCount].pFont=pFont;
	    _paSPS2UPBuffer[_iSPS2UPBufferCount].cCharacter=*pcCurrent;
	    
	    _iSPS2UPCurrentX+=iCharWidth;
	    _iSPS2UPBufferCount++;
	  }
	}
      }

      break;
    }

    pcCurrent++;
  }

  return iReturn;
}

/**
 * This function returns the height of the specified font. 
 * If pFont == 0 then the default font is used
 */
int sps2UPrintfFontHeight(sps2UFontStruct *pFont) {
  if (!pFont) {
    pFont=&_sps2UPFontFixed;
  }

  return pFont->iAscent+pFont->iDescent;
}

/**
 * Computes and returns the width of a given string for a specific font.
 * If pFont == 0 then the default font is used
 */
int sps2UPrintfStringWidth(sps2UFontStruct *pFont, const char *pcString) {
  int iWidth=0;

  if (!pFont) {
    pFont=&_sps2UPFontFixed;
  }

  while (*pcString) {
    iWidth+=_sps2UPCharWidth(pFont, *pcString);
    pcString++;
  }

  return iWidth;
}

/*****************************************************************************
 * SPS2 Printf private functions
 *****************************************************************************/
/**
 * Advances to the next line.
 */
void _sps2UPAdvanceLine(sps2UFontStruct *pFont) {
  _iSPS2UPCurrentX=_iSPS2UPStartX;
  _iSPS2UPCurrentY+=sps2UPrintfFontHeight(pFont);
}

/**
 * Our current and only font is fixed to 7 pixels width.
 */
int _sps2UPCharWidth(sps2UFontStruct *pFont, char cChar) {
  return pFont->aiWidths[(unsigned int) cChar];
}

/**
 * Outputs a single character.
 */
void _sps2UPOutputEntry(sps2UPrintfEntry *pEntry) {
  sps2GIFTag_t gifTag;
  sps2GIFPackedRegister_t gifRegister;
  int iRowLooper, iColLooper;
  unsigned int uiChar=(unsigned int) pEntry->cCharacter;

  int iFontHeight=pEntry->pFont->iAscent+pEntry->pFont->iDescent;
  int iFontWidth=pEntry->pFont->aiWidths[uiChar];

  int iPixel=0;

  while (*EE_GIF_STAT_OFF(_pvSPS2UEERegisters) & (31 << 24)); 

  // Prepare the GIF tag  
  gifTag.i128=0;       // Blank it out
  gifTag.s.NLOOP=pEntry->pFont->aiNonZeroBits[uiChar];
  gifTag.s.PRE=1;      // We are providing a valid PRIM value
  gifTag.s.EOP=1;      // End of packet -- draw me please!
  gifTag.s.PRIM=0;     // 0 = point
  gifTag.s.FLG=GIF_FLG_PACKED; // We're using the packed format for data
  gifTag.s.NREG=0;     // Start off with zero registers, increment as we
                         // populate the descriptions
  
  // Add a register.  It will be an RGBAQ register
  SPS2_SET_GIF_REG(gifTag, gifTag.s.NREG++, GIF_REG_RGBAQ);
  SPS2_SET_GIF_REG(gifTag, gifTag.s.NREG++, GIF_REG_XYZ2);

  // Send the GIF tag
  DPUT_EE_GIF_FIFO_OFF(_pvSPS2UEERegisters, gifTag.i128);

  for (iRowLooper=0;iRowLooper<iFontHeight;iRowLooper++) {
    int iRowValue=pEntry->pFont->auiCharacters[uiChar][iRowLooper];

    for (iColLooper=0;iColLooper<iFontWidth;iColLooper++) {
      if (iRowValue & (1 << iColLooper)) {
	gifRegister.i128=0;
	gifRegister.XYZ2.s.X=(iColLooper+(int) pEntry->sX+2048) << 4;
	gifRegister.XYZ2.s.Y=(iRowLooper+(int) pEntry->sY+2048) << 4;
	gifRegister.XYZ2.s.Z=pEntry->iZ;
	DPUT_EE_GIF_FIFO_OFF(_pvSPS2UEERegisters, pEntry->color);
	DPUT_EE_GIF_FIFO_OFF(_pvSPS2UEERegisters, gifRegister.i128);
	
	if ((++iPixel % 8) ==0)
	  while (*EE_GIF_STAT_OFF(_pvSPS2UEERegisters) & (31 << 24)); 
      }
    }
  }

  while (*EE_GIF_STAT_OFF(_pvSPS2UEERegisters) & (31 << 24)); 
}

/*****************************************************************************
 * SPS2 Non-Inline Functions
 *****************************************************************************/

/**
 * Non-inline counterpart of _sps2MapEERegisters.
 * See _sps2MapEERegisters documentation for more information.
 */
void *_sps2UNIMapEERegisters(void *pvWhere, int iDeviceHandle) {
  return _sps2MapEERegisters(pvWhere, iDeviceHandle);
}

/**
 * Non-inline counterpart of _sps2MapGSRegisters.
 * See _sps2MapGSRegisters documentation for more information.
 */
void *_sps2UNIMapGSRegisters(void *pvWhere, int iDeviceHandle) {
  return _sps2MapGSRegisters(pvWhere, iDeviceHandle);
}

/**
 * Non-inline counterpart of _sps2MapVUMemory.
 * See _sps2MapVUMemory documentation for more information.
 */
void *_sps2UNIMapVUMemory(void *pvWhere, int iDeviceHandle) {
  return _sps2MapVUMemory(pvWhere, iDeviceHandle);
}

/**
 * Non-inline counterpart of _sps2MapScratchPad.
 * See _sps2MapScratchPad documentation for more information.
 */
void *_sps2UNIMapScratchPad(void *pvWhere, int iDeviceHandle) {
  return _sps2MapScratchPad(pvWhere, iDeviceHandle);
}

/**
 * Non-inline counterpart of sps2Init.
 * See sps2Init documentation for more information.
 */
int sps2UNIInit() {
  return sps2Init();
}

/**
 * Non-inline counterpart of sps2Release.
 * See sps2Release documentation for more information.
 */
void sps2UNIRelease(int iDevice) {
  sps2Release(iDevice);
}


/**
 * Non-inline counterpart of _sps2Open.
 * See _sps2Open documentation for more information.
 */
int _sps2UNIOpen() {
  return _sps2Open();
}

/**
 * Non-inline counterpart of _sps2Close.
 * See _sps2Close documentation for more information.
 */
void _sps2UNIClose(int iDeviceHandle) {
  _sps2Close(iDeviceHandle);
}

/**
 * Non-inline counterpart of sps2FlushCache.
 * See sps2FlushCache documentation for more information.
 */
int sps2UNIFlushCache(int iDeviceHandle) {
  return sps2FlushCache(iDeviceHandle);
}


/**
 * Non-inline counterpart of sps2EnableCOP2Access.
 * See sps2EnableCOP2Access documentation for more information.
 */
int _sps2UNIEnableCOP2Access(int iDeviceHandle) {
  return _sps2EnableCOP2Access(iDeviceHandle);
}

/**
 * Non-inline counterpart of sps2GetPhysicalAddress.
 * See sps2GetPhysicalAddress documentation for more information.
 */
unsigned long sps2UNIGetPhysicalAddress(void *pvAddress,
					sps2Memory_t *pDescriptor) {
  return sps2GetPhysicalAddress(pvAddress, pDescriptor);
}


/**
 * Non-inline counterpart of sps2Allocate.
 * See sps2Allocate documentation for more information.
 */
sps2Memory_t *sps2UNIAllocate(unsigned long ulSize,
			      int iMapOptions,
			      int iDeviceHandle) {
  return sps2Allocate(ulSize, iMapOptions, iDeviceHandle);
}

/**
 * Non-inline counterpart of sps2Remap.
 * See sps2Remap documentation for more information.
 */
sps2Memory_t *sps2UNIRemap(sps2Memory_t *pOriginalArea,
			   int iMapOptions,
			   int iDeviceHandle) {
  return sps2Remap(pOriginalArea, iMapOptions, iDeviceHandle);
}

/**
 * Non-inline counterpart of sps2Free.
 * See sps2Free documentation for more information.
 */
void sps2UNIFree(sps2Memory_t *pMapping) {
  sps2Free(pMapping);
}

/**
 * Non-inline counterpart of sps2WaitForDMA.
 * See sps2WaitForDMA documentation for more information.
 */
int sps2UNIWaitForDMA(int iChannel,
		      int iDeviceHandle) {
  return sps2WaitForDMA(iChannel, iDeviceHandle);
}

/**
 * Non-inline counterpart of sps2SetOperationMode.
 * See sps2SetOperationMode documentation for more information.
 */
int sps2UNISetOperationMode(int iSupervisorMode,
			    int iDeviceHandle) {
  return sps2SetOperationMode(iSupervisorMode, iDeviceHandle);
}

/**
 * Non-inline counterpart of sps2SetEIDIEnabled.
 * See sps2SetEIDIEnabled documentation for more information.
 */
int _sps2UNISetEIDIEnabled(int iEnabled,
			    int iDeviceHandle) {
  return _sps2SetEIDIEnabled(iEnabled, iDeviceHandle);
}

/**
 * Non-inline counterpart of sps2ValidateDMA.
 * See sps2ValidateDMA documentation for more information.
 */
/*
int sps2UNIValidateDMA(int iDeviceHandle, int iChannel,
		       sps2uint32 ui32CHCR, int iSizeLimit) {
  return sps2ValidateDMA(iDeviceHandle, iChannel, ui32CHCR, iSizeLimit);
}
*/
