/**
 * x_font_to_sps2_u_font.c
 * Copyright 2004 Terratron Technologies Inc.  All rights reserved.
 * Author: Steven Osman
 * 2/11/2004
 */

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

void outputHeader(char *pcFontName);
void outputFooter();
int outputChar(Display *display, Drawable drawable, int x, int y, int w,
	       int h);
void outputAlphabet(Display *display, Drawable drawable, GC gc,
		    char *pcXFontName, char *pcFontName);

int main (int argc, char **argv) {
 Display         *pDisplay;
 Window          window;
 GC              gc;
 XEvent          event;
 XSizeHints      hint;
 unsigned long   ulWhite, ulBlack;
 int             screen;

 if (argc<3) {
   fprintf(stderr, "Usage\n%s <x font> <structure name>\n"
	   "\t<x font> is the name of a font (see xlsfonts)\n"
	   "\t<structure name> is the name of the structure to create\n",
	   argv[0]);
   exit(-1);
 }

 pDisplay = XOpenDisplay("");
 screen  = DefaultScreen(pDisplay);

 ulBlack = WhitePixel(pDisplay, screen);
 ulWhite = BlackPixel(pDisplay, screen);

 hint.x = 200; hint.y = 300;
 hint.width = 350; hint.height = 250;
 hint.flags = PPosition | PSize;

 window=XCreateSimpleWindow (pDisplay, DefaultRootWindow(pDisplay),
			     hint.x, hint.y, hint.width, hint.height,
			     5, ulWhite, ulBlack);

 gc = XCreateGC (pDisplay, window, 0, 0);

 XSetBackground (pDisplay, gc, ulBlack);
 XSetForeground (pDisplay, gc, ulWhite);
 XSelectInput (pDisplay, window, ExposureMask);
 XMapRaised (pDisplay, window);

 while (1) {
  XNextEvent (pDisplay, &event);
  if (event.type==Expose) {
    if (event.xexpose.count == 0) {
      outputAlphabet(pDisplay, window, gc, argv[1], argv[2]);
      break;
    }
  }
 }

 XFreeGC (pDisplay, gc);
 XDestroyWindow (pDisplay, window);
 XCloseDisplay (pDisplay);
 exit (0);
}
   
void outputAlphabet(Display *pDisplay, Drawable drawable, GC gc,
		    char *pcXFontName, char *pcFontName) {
  int iChar;
  char acString[2]={0,0};
  int iDirection, iAscent, iDescent;
  int iFirstChar=32, iLastChar=127;
  int aiWidths[256], aiNonZeroBits[256];
  XCharStruct overall;
  char acSpace[]=" ";
  Font font;

  font=XLoadFont(pDisplay, pcXFontName);
  XSetFont(pDisplay, gc, font);

  XQueryTextExtents(pDisplay, font, acSpace, strlen(acSpace),
		    &iDirection, &iAscent, &iDescent, &overall);
  memset(aiWidths, 0, sizeof(aiWidths));
  memset(aiNonZeroBits, 0, sizeof(aiNonZeroBits));


  /* Start file */
  outputHeader(pcFontName);

  /* Start characters */
  for (iChar=0;iChar<256;iChar++) {
    if (iChar>=iFirstChar && iChar<=iLastChar) {
      acString[0]=iChar;

      XQueryTextExtents(pDisplay, font, acString, strlen(acString),
			&iDirection, &iAscent, &iDescent, &overall);
      if (overall.width>32) {
	fprintf(stderr, 
		"Character %i is too wide. Its width %i is greater than 32.\n",
		iChar, overall.width);
	exit(-2);
      }

      XDrawImageString (pDisplay, drawable,
			gc, 50, 50, acString, strlen (acString));
      printf("static unsigned int _aui%s%i[]={\n", pcFontName, iChar);
      aiNonZeroBits[iChar]=outputChar(pDisplay, drawable, 50, 50-iAscent,
				      overall.width, iAscent+iDescent);
      printf("};\n");
      aiWidths[iChar]=overall.width;
    }
  }
  /* End characters */

  /* Start structure */
  printf("static sps2UFontStruct %s={\n"
	 "\t%i,\t/* First character */\n"
	 "\t%i,\t/* Last character */\n"
	 "\t%i,\t/* Ascent height */\n"
	 "\t%i,\t/* Descent height */\n", pcFontName, iFirstChar, iLastChar,
	 iAscent,
	 iDescent);

  /* Start character index */
  printf("\t{\n");
  for (iChar=0;iChar<256;iChar++) {
    if (iChar>=iFirstChar && iChar<=iLastChar) {
      printf("\t\t_aui%s%i,\n", pcFontName, iChar);
    } else {
      printf("\t\t(unsigned int *) 0,\n");
    }
  }
  printf("\t},\n");
  /* End character index */

  /* Start non-zero bits */
  printf("\t{\n");
  for (iChar=0;iChar<256;iChar++) {
    if (iChar % 8 == 0)
      printf("\t\t");

    printf("%4i, ", aiNonZeroBits[iChar]);

    if (iChar % 8 == 7)
      printf("\n");
  }
  printf("\t},\n");
  /* End non-zero bits */

  /* Start character widths */
  printf("\t{\n");

  for (iChar=0;iChar<256;iChar++) {
    if (iChar % 8 == 0)
      printf("\t\t");

    printf("%2i, ", aiWidths[iChar]);

    if (iChar % 8 == 7)
      printf("\n");
  }
  printf("\t},\n");

  /* End character widths */
	 
  printf("};\n");

  /* End structure */

  outputFooter();

  /* End file */
}

int outputChar(Display *pDisplay, Drawable drawable, int x, int y,
	       int w, int h) {
  XImage *pImage;
  int iXLooper, iYLooper;
  unsigned int uiLine;
  char acString[33];
  char cSymbol;
  int iNonZeroBits=0;

  pImage=XGetImage(pDisplay, drawable, x, y, w, h, -1, XYPixmap);
  memset(acString, 0, 33);

  for (iYLooper=0;iYLooper<h;iYLooper++) {
    uiLine=0;
    for (iXLooper=0;iXLooper<w;iXLooper++) {
      if (XGetPixel(pImage, iXLooper, iYLooper)==0) {
	uiLine|=1 << iXLooper;
	iNonZeroBits++;
	cSymbol='X';
      } else {
	cSymbol='.';
      }

      acString[iXLooper]=cSymbol;
    }
    printf("\t0x%08x, /* %s */\n", uiLine, acString);
  }
  XDestroyImage(pImage);

  return iNonZeroBits;
}

void outputHeader(char *pcFontName) {
  printf("#ifndef __FONT_%s\n"
	 "#define __FONT_%s\n"
	 "\n", pcFontName, pcFontName);

  printf("#ifndef __SPS2_U_FONT_STRUCT\n"
	 "#define __SPS2_U_FONT_STRUCT\n"
	 "\n"
	 "typedef struct {\n"
	 "\tint iFirstCharacter;\n"
	 "\tint iLastCharacter;\n"
	 "\tint iAscent;\n"
	 "\tint iDescent;\n"
	 "\tunsigned int *auiCharacters[256];\n"
	 "\tint aiNonZeroBits[256];\n"
	 "\tint aiWidths[256];\n"
	 "} sps2UFontStruct;\n"
	 "#endif\n"
	 "\n");
}

void outputFooter() {
  printf("#endif\n");
}
